==> org-lsst-ccs-groovy-description-5.3.12-sources.jar: org/lsst/ccs/description/groovy/GroovyComponentBuilder.java <==
package org.lsst.ccs.description.groovy;

import groovy.util.BuilderSupport;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

import java.util.Map;
import java.util.Set;
import org.lsst.ccs.description.ComponentNode;

/**
 * Class dedicated to building nodes from a groovy description file.
 * All it does is build DescriptiveNodes from the provided arguments.
 * It also sets the parent/child relationship as the tree is built.
 *
 * This class is overwritten by CCSBuilder in the startup package.
 * CCSBuilder is used directly in the groovy files.
 *
 * @author The LSST CCS Team
 */
class GroovyComponentBuilder extends BuilderSupport {

    private final List<ComponentNode> componentsList = new ArrayList<>();
        
    @Override
    public void setParent(Object o, Object o1) {
        GroovyComponentNode parent = (GroovyComponentNode) o;
        parent.addChild((GroovyComponentNode) o1);
    }

    @Override
    public Object createNode(Object name) {
        throw new IllegalStateException("class missing for node " + name);
    }

    @Override
    public Object createNode(Object name, Object value) {
        return createNode(name, Collections.EMPTY_MAP, value);
    }

    @Override
    public Object createNode(Object name, Map attributes) {
        throw new IllegalStateException("class missing for node " + name);
    }

    @Override
    public Object createNode(Object name, Map attributes, Object value) {
        String key = (String)name;
        try {
            GroovyComponentNode node = new GroovyComponentNode(key, attributes, (Class) value);
            getInternalDictionary().add(node);
            return node;
        } catch (Exception ex) {
            throw new RuntimeException("at node : " + name + ": " + ex.getMessage() , ex);
        }
    }

    protected GroovyComponentNode getCurrentNode() {
        GroovyComponentNode current = (GroovyComponentNode) getCurrent();
        return current;
    }
    
    private List<ComponentNode> getInternalDictionary() {
        return componentsList;
    }

    /**
     * A hook to allow nodes to be processed once they have had all of their children applied.
     * We use this hook to set the appropriate constructor to this node, given the list of attributes
     *
     * @param parent the parent of the node being processed
     * @param node the current node being processed
     */
    @Override
    public void nodeCompleted(Object parent, Object node) {
        GroovyComponentNode gNode = (GroovyComponentNode) node;

        try {
            gNode.attributes = (Map)resolveReferencesIn(gNode.attributes);
            Object component =  new GroovyConstructorInvoker().invokeConstructor(gNode.getCls(), gNode.getAttributes());            
            gNode.setComponent(component);
        } catch (Exception ex) {
            throw new RuntimeException("At node : " + gNode.getKey() + " : " + ex.getMessage(), ex);
        }
    }

    private Object resolveReferencesIn(Object obj) {
        if (obj == null) {
            return null;
        } else if(obj instanceof GroovyReferencePlaceHolder) {
            return resolveReference(null, (GroovyReferencePlaceHolder)obj);
        } else if(Map.class.isAssignableFrom(obj.getClass())) {
            for ( Object key : ((Map)obj).keySet() ) {
                Object value = ((Map)obj).get(key);
                if(value instanceof GroovyReferencePlaceHolder) {
                    ((Map)obj).put(key, resolveReference((String)key, (GroovyReferencePlaceHolder)value));
                } else {
                    ((Map)obj).put(key, resolveReferencesIn(value));
                }
            }
            return obj;
        } else if (obj.getClass().isArray()) {
            for(int i = 0; i<Array.getLength(obj); i++) {
                Array.set(obj, i, resolveReferencesIn(Array.get(obj,i)));
            }
            return obj;
        } else if (java.util.List.class.isAssignableFrom(obj.getClass())) {
            List resolvedList = new ArrayList<>();
            for (Object o : (List)obj) {
                resolvedList.add(resolveReferencesIn(o));
            }
            return resolvedList;
        } else if (java.util.Set.class.isAssignableFrom(obj.getClass())) {
            Set resolvedSet = new HashSet<>();
            for (Object o : (Set)obj) {
                resolvedSet.add(resolveReferencesIn(o));
            }
            return resolvedSet;
        } else {
            return obj;
        }
    }
    
    private Object resolveReference(String key, GroovyReferencePlaceHolder ph) {
        String ref = ph.getReference() == null ? key : ph.getReference();               
        if (ref == null) {
            throw new RuntimeException("cannot infer reference name");
        }
        //First try to look by key. If that does not work, try to look for path.
        List<ComponentNode> matchingComponents = getComponentsForKey(ref);        
        if ( matchingComponents.size() != 1 ) {
            matchingComponents = getComponentsForPath(ref);        
        }
        if ( matchingComponents.size() != 1 ) {
            throw new RuntimeException("Found "+matchingComponents.size()+" components for key "+ref+"!!! There must be only one!!!");
        }
        ComponentNode compNode = matchingComponents.get(0);
        if ( compNode != null ) {               
            return compNode.getComponent();
        } else {
            throw new RuntimeException("null reference associated to " + ref);
        }
    }
    
    private List<ComponentNode> getComponentsForPath(String path) {
        List<ComponentNode> result = new ArrayList<>();
        for ( ComponentNode node : getInternalDictionary() ) {
            if ( node.getPath().equals(path) ) {
                result.add(node);
            }
        }
        return result;
    }
    private List<ComponentNode> getComponentsForKey(String key) {
        List<ComponentNode> result = new ArrayList<>();
        for ( ComponentNode node : getInternalDictionary() ) {
            if ( node.getKey().equals(key) ) {
                result.add(node);
            }
        }
        return result;
    }
    
}

==> org-lsst-ccs-groovy-description-5.3.22-sources.jar: org/lsst/ccs/description/groovy/GroovyComponentBuilder.java <==
package org.lsst.ccs.description.groovy;

import groovy.util.BuilderSupport;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

import java.util.Map;
import java.util.Set;
import org.lsst.ccs.description.ComponentNode;

/**
 * Class dedicated to building nodes from a groovy description file.
 * All it does is build DescriptiveNodes from the provided arguments.
 * It also sets the parent/child relationship as the tree is built.
 *
 * This class is overwritten by CCSBuilder in the startup package.
 * CCSBuilder is used directly in the groovy files.
 *
 * @author The LSST CCS Team
 */
class GroovyComponentBuilder extends BuilderSupport {

    private final List<ComponentNode> componentsList = new ArrayList<>();
    private ComponentNode topNode = null;
            
    @Override
    public void setParent(Object o, Object o1) {
        GroovyComponentNode parent = (GroovyComponentNode) o;
        parent.addChild((GroovyComponentNode) o1);
    }

    @Override
    public Object createNode(Object name) {
        throw new IllegalStateException("class missing for node " + name);
    }

    @Override
    public Object createNode(Object name, Object value) {
        return createNode(name, Collections.EMPTY_MAP, value);
    }

    @Override
    public Object createNode(Object name, Map attributes) {
        throw new IllegalStateException("class missing for node " + name);
    }

    @Override
    public Object createNode(Object name, Map attributes, Object value) {
        String key = (String)name;
        try {
            GroovyComponentNode node = new GroovyComponentNode(key, attributes, (Class) value);
            if ( getInternalDictionary().isEmpty() ) {
                topNode = node;
            }            
            getInternalDictionary().add(node);
            return node;
        } catch (Exception ex) {
            throw new RuntimeException("at node : " + name + ": " + ex.getMessage() , ex);
        }
    }

    protected GroovyComponentNode getCurrentNode() {
        GroovyComponentNode current = (GroovyComponentNode) getCurrent();
        return current;
    }
    
    private List<ComponentNode> getInternalDictionary() {
        return componentsList;
    }

    /**
     * A hook to allow nodes to be processed once they have had all of their children applied.
     * We use this hook to set the appropriate constructor to this node, given the list of attributes
     *
     * @param parent the parent of the node being processed
     * @param node the current node being processed
     */
    @Override
    public void nodeCompleted(Object parent, Object node) {
        GroovyComponentNode gNode = (GroovyComponentNode) node;
        try {
            gNode.attributes = (Map)resolveReferencesIn(gNode.attributes);
            String nodeName = gNode.getPath();
            if ( nodeName.isEmpty() ) {
                nodeName = gNode.getKey();
            }
            Object component =  new GroovyConstructorInvoker().invokeConstructor(nodeName, gNode.getCls(), gNode.getAttributes(), topNode);            
            gNode.setComponent(component);
        } catch (Exception ex) {
            throw new RuntimeException("At node : " + gNode.getKey() + " : " + ex.getMessage(), ex);
        }
    }

    private Object resolveReferencesIn(Object obj) {
        if (obj == null) {
            return null;
        } else if(obj instanceof GroovyReferencePlaceHolder) {
            return resolveReference(null, (GroovyReferencePlaceHolder)obj);
        } else if(Map.class.isAssignableFrom(obj.getClass())) {
            for ( Object key : ((Map)obj).keySet() ) {
                Object value = ((Map)obj).get(key);
                if(value instanceof GroovyReferencePlaceHolder) {
                    ((Map)obj).put(key, resolveReference((String)key, (GroovyReferencePlaceHolder)value));
                } else {
                    ((Map)obj).put(key, resolveReferencesIn(value));
                }
            }
            return obj;
        } else if (obj.getClass().isArray()) {
            for(int i = 0; i<Array.getLength(obj); i++) {
                Array.set(obj, i, resolveReferencesIn(Array.get(obj,i)));
            }
            return obj;
        } else if (java.util.List.class.isAssignableFrom(obj.getClass())) {
            List resolvedList = new ArrayList<>();
            for (Object o : (List)obj) {
                resolvedList.add(resolveReferencesIn(o));
            }
            return resolvedList;
        } else if (java.util.Set.class.isAssignableFrom(obj.getClass())) {
            Set resolvedSet = new HashSet<>();
            for (Object o : (Set)obj) {
                resolvedSet.add(resolveReferencesIn(o));
            }
            return resolvedSet;
        } else {
            return obj;
        }
    }
    
    private Object resolveReference(String key, GroovyReferencePlaceHolder ph) {
        String ref = ph.getReference() == null ? key : ph.getReference();               
        if (ref == null) {
            throw new RuntimeException("cannot infer reference name");
        }
        //First try to look by key. If that does not work, try to look for path.
        List<ComponentNode> matchingComponents = getComponentsForKey(ref);        
        if ( matchingComponents.size() != 1 ) {
            matchingComponents = getComponentsForPath(ref);        
        }
        if ( matchingComponents.size() != 1 ) {
            throw new RuntimeException("Found "+matchingComponents.size()+" components for key "+ref+"!!! There must be only one!!!");
        }
        ComponentNode compNode = matchingComponents.get(0);
        if ( compNode != null ) {               
            return compNode.getComponent();
        } else {
            throw new RuntimeException("null reference associated to " + ref);
        }
    }
    
    private List<ComponentNode> getComponentsForPath(String path) {
        List<ComponentNode> result = new ArrayList<>();
        for ( ComponentNode node : getInternalDictionary() ) {
            if ( node.getPath().equals(path) ) {
                result.add(node);
            }
        }
        return result;
    }
    private List<ComponentNode> getComponentsForKey(String key) {
        List<ComponentNode> result = new ArrayList<>();
        for ( ComponentNode node : getInternalDictionary() ) {
            if ( node.getKey().equals(key) ) {
                result.add(node);
            }
        }
        return result;
    }
    
}
